/*
 * Copyright (c) 2007 Jan Hudecek
 * Copyright (c) 2006 Jakub Jermar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/** @addtogroup sync
 * @{
 */
/** @file rcu.h
 * @brief declarations for RCU
 */

#ifndef KERN_RCU_H_
#define KERN_RCU_H_

#include <arch/types.h>
#include <proc/tasklet.h>
#include <arch/barrier.h>
#include <arch.h>
#include <preemption.h>


/** Structure for callbacks */
typedef struct rcu_callback_list {
	/** next in the list */
	struct rcu_callback_list* next;
	/** pointer to callback function */
	void (*func)(void*);
	/** argument to pass to the callback */
	void* data;
} rcu_callback_list_t;


/** Read lock for RCU protected pointer */
#define rcu_read_lock() preemption_disable()

/** Release of read lock for RCU protected pointer */
#define rcu_read_unlock() preemption_enable()

/** Dereferencing of an RCU protected pointer */
#define rcu_dereference_pointer(p) (*(p))

/** Assigning a value to an RCU protected pointer */
#define rcu_assign_pointer(p, newp) {write_barrier(); (p)=(newp);}

/** RCU sync callback for those who don't need custom allocation */
#define rcu_sync_callback_normal_alloc(func, data) {\
	rcu_callback_list_t* rd = malloc(sizeof(rcu_callback_list_t),0);\
	rcu_sync_callback(func, data, rd);}

/** RCU sync callback without custom allocation, default callback (just frees the argument) */
#define rcu_sync_callback_normal_alloc_free(data) rcu_sync_callback_normal_alloc(&rcu_callback_free, data)

void rcu_init(void);
void rcu_synchronize(void);
void rcu_synchronize_callback_function(void* waitq);
void rcu_sync_callback(void (*func)(void* data), void* data, rcu_callback_list_t* rcu_struct);
void rcu_tasklet(void* data);	
void rcu_passQS(void);
void rcu_run_callbacks(void);
void rcu_callback_free(void* pointer);




#endif
